#
# Debugging:
#
# To use the pdb debugger on code in this file, you will need to disable
# the parts of NUU that capture output for the Log window.  Do that in
# two places.  In YumIO.c, change:
#    #undef DEBUG_DISABLE_CAPTURERS
#       =>
#    #define DEBUG_DISABLE_CAPTURERS
# Then just insert the following wherever you need to gain control with pdb:
#    debug()
# If stdin, stdout & stderr still give you troubles, you also could change this
# at the end of this file:
#    #sys.stdout = YumIOStdout()
#    #sys.stdin = YumIOStdin()
#

import sys
import os
sys.path.insert(0, '/usr/share/yum-cli')
sys.path.insert(0, '%s/python' % os.getenv('NUUSHARE', '/usr/share/nuu'))

def debug():
    import sys
    old_stdout = sys.stdout
    sys.stdout = sys.__stdout__
    import pdb
    pdb.set_trace()
    sys.stdout = old_stdout

# If the garbage collector is enabled (the default), if anyone does an
# update, presses Cancel, does it again, and presses OK, the update will
# complete successfully, but then the re-initialization will fail with an
# assertion:
#    Modules/gcmodule.c:378: delete_garbage:
#    Assertion `((((PyGC_Head *)(op)-1))->gc.gc_refs >= 0)' failed.
# I went so far as disabling every single Py_DECREF in the C++ code and
# verified that there were Py_INCREF's before the calls to PyTuple_SetItem
# and PyList_SetItem, just to make sure I wasn't leaving any dangling
# references, but the gc still blew up.  I suspect this is a bug in Python's
# gc, so I'm disabling it.
import gc
gc.disable()

# For versions of Python that do GIL locking, NUU must act as a multi-threaded
# application (even though NUU itself guarantees that python cannot be running
# in more than one thread at a time).  But the YumBaseCli.doTransaction will
# attempt to call signal.signal, and that is not allowed in any non-main thread.
# To workaround this, create a new signal pseudo-module (it really is a class
# instance) that mostly works like the original signal module, but where the
# signal function is overridden to do no real work.
import signal
__signal__ = signal
class Signal:
    pass
for meth in dir(__signal__):
    setattr(Signal, meth, getattr(__signal__, meth))
Signal.signal = staticmethod(lambda num, obj: __signal__.getsignal(num))
sys.modules['signal'] = Signal()
import signal

# In python 2.5, pysqlite has instituted a check that the sqlite object only can
# be accessed by a single python thread.  This is intended to catch common
# errors.  But NUU is careful to ensure that no two threads will access the sql
# data at the same time, so this check is unnecessary.  Rather than redesign NUU
# to workaround an unnecessary check, disable it by overriding
# sqlitecache.RepodataParserSqlite.open_database and making it pass
# check_same_thread=0.
#
try:
   import sqlitecachec
   __sqlitecachec__ = sqlitecachec
   class RepodataParserSqlite(__sqlitecachec__.RepodataParserSqlite):
       def open_database(self, filename):
           if not filename:
               return None
           con = __sqlitecachec__.sqlite.connect(filename, check_same_thread=0)
           if __sqlitecachec__.sqlite.version_info[0] > 1:
               con.row_factory = __sqlitecachec__.sqlite.Row
           return con

   class Sqlitecachec:
       pass
   for meth in dir(__sqlitecachec__):
       setattr(Sqlitecachec, meth, getattr(__sqlitecachec__, meth))
   Sqlitecachec.RepodataParserSqlite = RepodataParserSqlite
   sys.modules['sqlitecachec'] = Sqlitecachec()
   import sqlitecachec
except ImportError:
   pass

#
# This is used to override RPMInstallCallback, which reports on progress for
# installs, upgrades, erases, etc.
#
# yum-3.2.2 and before:
#
# The YumBaseCli is organized pretty well with the ability to override I/O
# methods in mind.  But they dropped the ball with the
# callback.RPMInstallCallback class.  It's created and visible only within
# YumBaseCli.doTransaction, so there's no clean way to override it.
# Furthermore, it writes directly to stdout!
#
# One approach to resolve this is to override sys.stdout and parse its output.
# But this creates dependencies on exact text from yum, and so it would be
# likely to break with future yum versions.
#
# Another approach is to override the whole RPMInstallCallback class and
# replicate all its code.  But this also creates a dependency on the particular
# version of yum from which the RPMInstallCallback is replicated.
#
# So, instead this code overrides the RPMInstallCallback class, and creates a
# callback pseudo-module (it really is a class instance) where the overriding
# class is inserted to do the real work.  It then calls the original callback
# as __callback__ where appropriate.
#
# yum-3.2.3 and after:
#
# The organization is different, but equally hard to alter from outside the
# YumBaseCli.  In this case, the class that must be overridden is the
# output.YumCliRPMCallback class.  It is handled with a new output pseudo-module
# (really a class instance) which overrides the YumCliRPMCallback.
#
# Note that we get no callback for the transaction test (output=0) in this case.
#
import callback
import yumIOC
__callback__ = callback
class RPMInstallCallback(__callback__.RPMInstallCallback):
    def __init__(self, output=1):
        __callback__.RPMInstallCallback.__init__(self)
        yumIOC.tsInitCallback(output)

    def _localprint(self, msg):
        if self.output:
            yumIOC.log(0, msg)

    def callback(self, what, bytes, total, h, user):
        yumIOC.tsCallback(what, bytes, total, h, user)
        # Starting around yum 3.0, the _localprint function was eliminated and
        # the callback code started writing directly to stdout.  It's necessary
        # to intercept that, and that is done with the YumIOStdout.sendToLog
        # flag.
        if hasattr(sys.stdout, 'sendToLog'): sys.stdout.sendToLog = 1
        result = __callback__.RPMInstallCallback.callback(self, what, bytes,
                                                          total, h, user)
        if hasattr(sys.stdout, 'sendToLog'): sys.stdout.sendToLog = 0
        return result
    
class Callback:
    pass
for meth in dir(__callback__):
    setattr(Callback, meth, getattr(__callback__, meth))
Callback.RPMInstallCallback = RPMInstallCallback
sys.modules['callback'] = Callback()
import callback

# rsyslog-5.8.10 malfunctions if passed UTF-8-encoded messages.  That happens if
# yum package names are in unicode.  And that happens in NUU because package
# names are taken from early collect passes (where they are unicode) and are fed
# back into later collect passes.  (It might be possible to finagle this with
# straight yum too, but I didn't try very hard.)
#
# This use of unicode is no problem for NUU itself.  But the syslog messages
# that python produces for unicode message strings confuse rsyslog-5.8.10 into
# thinking that the LOG_USER+LOG_INFO (14) messages really are
# LOG_KERN+LOG_EMERG (0) and it freaks out and yells at everyone on the system
# for every operation.
#
# To workaround this, I override the logging.handlers.SysLogHandler class such
# that it behaves as much like its normal self as possible, but it has an
# overriding format() function which always changes the formatted message into a
# simple string instead of potentially being unicode.
try:
   import logging.handlers
   __handlers__ = logging.handlers
   class SysLogHandler(__handlers__.SysLogHandler):
       def format(self, record):
           return __handlers__.SysLogHandler.format(self, record).encode('ascii', 'replace')

   import sys,imp
   Handlers = imp.new_module('Handlers')
   for meth in dir(__handlers__):
       if meth.startswith('__'):
           pass
       else:
           setattr(Handlers, meth, getattr(__handlers__, meth))
   Handlers.SysLogHandler = SysLogHandler
   logging.handlers = Handlers
   sys.modules['logging.handlers'] = Handlers
   import logging.handlers
except ImportError:
   pass

# pycurl cannot handle unicode URLs.  Override urlgrabber.grabber.URLGrabber's
# urlgrab functions with versions that transform any input unicode strings into
# ascii strings, much like for SysLogHandler above.
try:
   import urlgrabber.grabber
   __grabber__ = urlgrabber.grabber
   class URLGrabber(__grabber__.URLGrabber):
       def urlgrab(self, url, filename=None, **kwargs):
           return __grabber__.URLGrabber.urlgrab(self, url.encode('ascii', 'replace'), filename, **kwargs)

   import sys,imp
   Grabber = imp.new_module('Grabber')
   for meth in dir(__grabber__):
       if meth.startswith('__'):
           pass
       else:
           setattr(Grabber, meth, getattr(__grabber__, meth))
   Grabber.URLGrabber = URLGrabber
   urlgrabber.grabber = Grabber
   sys.modules['urlgrabber.grabber'] = Grabber
   import urlgrabber.grabber
except ImportError:
   pass

# In Yum 2.4.x and higher, repos were reimplemented with a new sqlite layer.
# Each repo has a YumSqlitePackageSack.  Unfortunately, that class has a
# new implementation of addDict that never calls the callback, despite being
# passed it.  This class will be derived from YumSqlitePackagSack and will add
# an addDict function that does call the callback when appropriate.
# Objects of this class will be created in setupRepos and will be used to
# replace the sacks for the repos at that time.
try:
   import yum.sqlitesack
   class YumIOSqlitePackageSack(yum.sqlitesack.YumSqlitePackageSack):
       def __init__(self, original):
           self.__dict__ = original.__dict__.copy()
       def addDict(self, repo, datatype, dataobj, callback=None):
           if not self.added.has_key(repo) or datatype not in self.added[repo]:
               if callback: callback.progressbar(-1, -1, str(repo))
           yum.sqlitesack.YumSqlitePackageSack.addDict(self, repo, 
                                                       datatype, dataobj,
                                                       callback)
except ImportError:
   pass

try:
   import output
   import yumIOC
   __output__ = output
   class YumCliRPMCallBack(__output__.YumCliRPMCallBack):
       def __init__(self, ayum=None):
           __output__.YumCliRPMCallBack.__init__(self)
           yumIOC.tsInitCallback(1)

       def event(self, package, action, te_current, te_total, ts_current, ts_total):
           yumIOC.tsEvent(package, action,
                          te_current, te_total,
                          ts_current, ts_total)
           # Starting around yum 3.0, the _localprint function was eliminated and
           # the output code started writing directly to stdout.  It's necessary
           # to intercept that, and that is done with the YumIOStdout.sendToLog
           # flag.
           if hasattr(sys.stdout, 'sendToLog'): sys.stdout.sendToLog = 1
           result = __output__.YumCliRPMCallBack.event(self, package, action,
                                                       te_current, te_total,
                                                       ts_current, ts_total)
           if hasattr(sys.stdout, 'sendToLog'): sys.stdout.sendToLog = 0
           return result

   class Output:
       pass
   for meth in dir(__output__):
       setattr(Output, meth, getattr(__output__, meth))
   Output.YumCliRPMCallBack = YumCliRPMCallBack
   sys.modules['output'] = Output()
   import output
except AttributeError:
   pass

#
# Main body of YumIO
#

import locale
import cli
import output
import yumIOC
import yum.transactioninfo
import yum.Errors
import rpm
import distutils.version

def yumVersion():
    if hasattr(yum, '__version__'):
        return distutils.version.StrictVersion(yum.__version__)
    elif hasattr(cli, '__version__'):
        return distutils.version.StrictVersion(cli.__version__)
    else:
        return distutils.version.StrictVersion('0.0')

def hasDistribGroups():
    return yumVersion() >= '2.5'

def usesPythonLogging():
    return yumVersion() >= '2.9.2'

def hasUpdatedTransactions():
    return yumVersion() >= '3.1.3'

# Yum 2.4 used its own yumsqlitecache.
# Yum 3.2 uses python's sqlitecachec.
# Either is acceptable and means that sqlite is present and therefore used.
try:
    import yum.sqlitecache
    import yum.sqlitesack
except ImportError:
    try:
        import sqlitecachec
        import yum.sqlitesack
    except ImportError:
        _usesSqlite = False
    else:
        _usesSqlite = True
else:
    _usesSqlite = True

def usesSqlite():
    return _usesSqlite

if usesPythonLogging():
    import logging

# Used to override sys.stdin so that it will claim that it isatty, which
# convinces a yum check that it is not running unattended.
__stdin__ = sys.stdin
class YumIOStdin:
    def __init__(self):
        pass
    def isatty(self):
        return 1
for meth in dir(__stdin__):
    if not hasattr(YumIOStdin, meth):
        setattr(YumIOStdin, meth, getattr(__stdin__, meth))

# Used to override sys.stdout
class YumIOStdout:
    def __init__(self):
        self.sendToLog = 0
        self.logStash = ''
        
    def write(self, msg):
        if self.sendToLog:
            msg = self.logStash + msg
            pieces = msg.split('\n')
            self.logStash = pieces[-1]
            del pieces[-1]
            for piece in pieces:
                yumIOC.log(0, piece)
        else:
            if self.logStash != '':
                yumIOC.log(0, self.logStash)
                self.logStash = ''
            yumIOC.stdoutWrite(msg)

    def isatty(self):
        return 0

    def flush(self):
        pass

# Used as the callback for collect progress for Conversions and Reads.
class YumIOCollectProgress:
    def progressbar(self, current, total, name=None):
        return yumIOC.collectProgress(current, total, name)

    # This is so the object can be called directly.  While newer versions of
    # yum (e.g. 2.4.2) call this object's progressbar function, older
    # versions of yum (e.g. 2.2.2) expect to be able to call this object
    # directly.
    __call__ = progressbar

    def log(self, value, msg):
        return yumIOC.log(value, msg)

# Used as the progressBar for repos, which reports on progress for downloads
# This replaces a YumTextMeter.
class YumIOCollectGrabMeter:
    def __init__(self):
        pass
        
    def start(self,
              filename = None,
              url      = None,
              basename = None,
              size     = None,
              now      = None,
              text     = None):
        return yumIOC.collectGrabMeterStart(self, filename, url, basename, size, text)
        
    def update(self, amount_read, now=None):
        return yumIOC.collectGrabMeterUpdate(self, amount_read)

    def end(self, amount_read, now=None):
        return yumIOC.collectGrabMeterEnd(self, amount_read)

    def failure(self, message, now=None):
        return yumIOC.collectGrabMeterFailure(self, message)

# Used as the progressBar for repos, which reports on progress for downloads
# This replaces a YumTextMeter.
class YumIOCollectGrabMultiMeter:
    # This is a workaround for when grabber.py breaks into the meter and mucks
    # around with the self.re.total value.  Make sure there is one.
    class RE:
       def __init__(self):
           self.total = 0

    def __init__(self):
        self.re = YumIOCollectGrabMultiMeter.RE()
        
    def start(self, total_files, total_size):
        return yumIOC.collectGrabMultiMeterStart(total_files, total_size)

    def newMeter(self):
        result = YumIOCollectGrabMeter()
        yumIOC.collectGrabMultiMeterNew(result)
        return result

    def removeMeter(self, meter):
        yumIOC.collectGrabMultiMeterRemove(meter)

    def end(self):
        return yumIOC.collectGrabMultiMeterEnd()

# Used as the progressBar for repos, which reports on progress for downloads
# This replaces a YumTextMeter.
class YumIOApplyGrabMeter:
    def __init__(self):
        pass
        
    def start(self,
              filename = None,
              url      = None,
              basename = None,
              size     = None,
              now      = None,
              text     = None):
        return yumIOC.applyGrabMeterStart(self, filename, url, basename, size, text)
        
    def update(self, amount_read, now=None):
        return yumIOC.applyGrabMeterUpdate(self, amount_read)

    def end(self, amount_read, now=None):
        return yumIOC.applyGrabMeterEnd(self, amount_read)

    def failure(self, message, now=None):
        return yumIOC.applyGrabMeterFailure(self, message)

# Used as the progressBar for repos, which reports on progress for downloads
# This replaces a YumTextMeter.
class YumIOApplyGrabMultiMeter:
    # This is a workaround for when grabber.py breaks into the meter and mucks
    # around with the self.re.total value.  Make sure there is one.
    class RE:
       def __init__(self):
           self.total = 0

    def __init__(self):
        self.re = YumIOApplyGrabMultiMeter.RE()
        
    def start(self, total_files, total_size):
        return yumIOC.applyGrabMultiMeterStart(total_files, total_size)

    def newMeter(self):
        result = YumIOApplyGrabMeter()
        yumIOC.applyGrabMultiMeterNew(result)
        return result

    def removeMeter(self, meter):
        yumIOC.applyGrabMultiMeterRemove(meter)

    def end(self):
        return yumIOC.applyGrabMultiMeterEnd()

# Used to override dsCallback by YumIO.__init__, which reports on progress for 
# dependency resolution
class YumIODsCallback(output.DepSolveProgressCallBack):
    def __init__(self, log, errorlog):
        if usesPythonLogging():
            output.DepSolveProgressCallBack.__init__(self)
        else:
            output.DepSolveProgressCallBack.__init__(self, log, errorlog)

    def tscheck(self):
        yumIOC.dsTsCheck()
        return output.DepSolveProgressCallBack.tscheck(self)
        
    def restartLoop(self):
        yumIOC.dsRestartLoop()
        return output.DepSolveProgressCallBack.restartLoop(self)
    
if usesPythonLogging():
    import yum.logginglevels
    
    debuglevel = {}
    last_ll = 51
    for dl in range(10):
        this_ll = yum.logginglevels.logLevelFromDebugLevel(dl)
        for ll in range(this_ll, last_ll):
            debuglevel[ll] = dl
        last_ll = this_ll
        for ll in range(0, last_ll):
            debuglevel[ll] = 10
        
    class FunctionHandler(logging.Handler):
        def __init__(self, function):
            logging.Handler.__init__(self)
            self.function = function
            self.formatter = None

        def emit(self, record):
            try:
                msg = self.format(record)
                self.function(debuglevel[record.levelno], msg)
            except:
                self.handleError(record)

# Used as a base class for YumIO to intercept callbacks and pass them to YumIOC
class YumIOC:

    # Used to override Loggers in YumIO.__init__
    
    def log(self, value, msg):
        return yumIOC.log(value, msg)

    def errorlog(self, value, msg):
        return yumIOC.errorlog(value, msg)

    # Used to override progress bar in YumIO.__init__
    
    # Overrides YumOutput
    
    def userconfirm(self,prompt=None,extra=None):
        return yumIOC.userconfirm()
    
    def reportDownloadSize(self,packages,install_only=True):
        yumIOC.reportDownloadSize(packages)
        return output.YumOutput.reportDownloadSize(self,packages)
            
    #
    # Versions of yum 2.2.x - 3.2.28 report Obsoletes incorrectly at least some
    # of the time.  This code detects and adds notification of any Obsoletes
    # that will happen as the result of installs or updates, including any
    # caused by dependency resolution.
    #
    # This is unnecessary as of 3.2.29, when
    # yum.transactioninfo.TransactionData.findObsoletedByThisMember was added.
    #
    def addObsoletes(self):
        if yumVersion() < '3.2.29':
            if yumVersion() >= '2.3.2':
                self.tsInfo.makelists()
                updated      = self.tsInfo.updated
                installed    = self.tsInfo.installed
                removed      = self.tsInfo.removed
                obsoleted    = self.tsInfo.obsoleted
                depupdated   = self.tsInfo.depupdated
                depinstalled = self.tsInfo.depinstalled
                depremoved   = self.tsInfo.depremoved
            else:
                (updated, installed, removed, obsoleted,
                 depupdated, depinstalled, depremoved) = self.tsInfo.makelists()
            for obsoletingMbr in updated + installed + depupdated + depinstalled:
                if self.up.obsoletes.has_key(obsoletingMbr.pkgtup):
                    for obsoletedTuple in self.up.obsoletes[obsoletingMbr.pkgtup]:
                        obsoletingPkg = self.getPackageObject(obsoletingMbr.pkgtup)
                        if yumVersion() >= '2.9.5':
                            (n, a, e, v, r) = obsoletedTuple
                            obsoletedHdr = self.rpmdb.searchNevra(name=n, arch=a, epoch=e, ver=v, rel=r)[0].hdr
                        else:
                            obsoletedHdr = self.rpmdb.returnHeaderByTuple(obsoletedTuple)[0]
                        obsoletedPkg = yum.packages.YumInstalledPackage(obsoletedHdr)
                        self.tsInfo.addObsoleting(obsoletingPkg, obsoletedPkg)
                        self.tsInfo.addObsoleted(obsoletedPkg, obsoletingPkg)

    def listTransaction(self):
        self.addObsoletes()
        yumIOC.listTransaction(self)
        return output.YumOutput.listTransaction(self)

    def postTransactionOutput(self):
        yumIOC.postTransactionOutput(self)
        return output.YumOutput.postTransactionOutput(self)

    def gpgsigcheck(self, pkgs):
        yumIOC.gpgsigcheck()
        return cli.YumBaseCli.gpgsigcheck(self, pkgs)

import rpmUtils.arch

class YumIO(YumIOC, cli.YumBaseCli):
    def __init__(self):
        locale.setlocale(locale.LC_ALL, '')
        cli.YumBaseCli.__init__(self)
        if usesPythonLogging():
            # The getOptionsConfig call will add a handler.  If this is a
            # re-initialization, we want to get rid of the old one first to
            # avoid duplicated output.
            logging.getLogger('yum.verbose').handlers = []
            logging.getLogger('yum').handlers = []
        # Read options.  'info' is a stunt to workaround YumBaseCli
        # command-line option processing.
        self.getOptionsConfig([ 'makecache', 'fast' ])
        # Get rid of Loggers from this class (if they exist, as they did in
        # roughly yum 2.5 and before) and use versions from YumIOC
        self.log = 0
        del self.log      
        self.errorlog = 0
        del self.errorlog
        if usesPythonLogging():
            verbose = logging.getLogger('yum.verbose')
            verbose.handlers = []
            verbose.addHandler(FunctionHandler(self.log))
            logger = logging.getLogger('yum')
            logger.handlers = []
            logger.addHandler(FunctionHandler(self.errorlog))
        if usesPythonLogging():
            # Ensure this logger reports TSINFO lines.
            self.verbose_logger.setLevel(logging.DEBUG)
            
        # Keep self.filelog so that /var/log/yum.log (or whatnot) is updated
        # Use progress bars from YumIOC
        self.repos.callback = YumIOCollectProgress()
        #if usesPythonLogging():
        #    callback = logging.getLogger('yum.filelogging.RPMInstallCallback')
        #    #callback.removeHandler(callback.handlers[0])
        #    callback_handler = FunctionHandler(self.log)
        #    callback.addHandler(callback_handler)
        self.repos.setProgressBar(None) # Overidden for appropriate operations
        self.dsCallback = None # Overidden during apply
        # Initialize biarch dict
        # Tolerate GPG keys which have empty arches
        self.biarch = { '': '' } 
        canonarch = rpmUtils.arch.getCanonArch()
        for arch in rpmUtils.arch.getArchList(canonarch):
            if arch != "noarch":
                self.biarch[arch] = canonarch
        if rpmUtils.arch.isMultiLibArch(canonarch):
            biarch = rpmUtils.arch.getMultiArchInfo(canonarch)[0]
            for arch in rpmUtils.arch.getArchList(biarch):
                if arch != "noarch":
                    self.biarch[arch] = biarch

    # Utility functions

    def lock(self):
        self.doLock('/var/run/yum.pid')

    def unlock(self):
        self.closeRpmDB()
        # yum 2.2.2's closeRpmDB will delete the rpmdb, but will leave the
        # localdbimported set to true, so that any future re-created rpmdb
        # won't be populated properly.  Workaround it.
        if hasattr(self, 'localdbimported'):
            self.localdbimported = 0
        self.doUnlock('/var/run/yum.pid')

    def usesPresto(self):
        try:
            return self.plugins._plugins.has_key('presto')
        except AttributeError:
            return False

    def hasMultiProgress(self):
        return hasattr(self.repos.repos.itervalues().next(), 'multi_callback')

    def setupRepos(self):
        # Simplified version of doRepoSetup that just lets exceptions
        # propagate to the caller, instead of calling sys.exit.
        if self.hasMultiProgress():
            self.repos.setProgressBar(YumIOCollectGrabMeter(),
                                      YumIOCollectGrabMultiMeter())
        else:
            self.repos.setProgressBar(YumIOCollectGrabMeter())
        try:
            self.log(2, 'Setting up Repos')
            for repo in self.repos.listEnabled():
                # Replace repo's package sack with one that actually does callbacks.
                # See YumIOSqlitePackageSack above for more details.  The exact
                # location where the sack is stashed varies depending on yum
                # version.
                if usesSqlite():
                    if hasattr(repo, '_sack'):
                        repo._getSack()
                        repo._sack = YumIOSqlitePackageSack(repo._sack)
                    elif hasattr(repo, 'sack'):
                        repo.sack = YumIOSqlitePackageSack(repo.sack)
                    elif hasattr(self.repos, 'pkgSack') \
                         and isinstance(self.repos.pkgSack,
                                        yum.sqlitesack.YumSqlitePackageSack):
                        self.repos.pkgSack = YumIOSqlitePackageSack(self.repos.pkgSack)
                if yumVersion() >= '2.9.6':
                    repo.setup(self.conf.cache)
                else:
                    if yumVersion() >= '2.5':
                        repo.cache = self.conf.cache
                    else:
                        repo.cache = self.conf.getConfigOption('cache')
                    repo.baseurlSetup()
                    repo.dirSetup()
                    self.log(3, 'Baseurl(s) for repo: %s' % repo.urls)
                    repo.getRepoXML(text=repo)
            self.log(2, 'Reading repository metadata in from local files')
            self.doSackSetup()
            self.repos.setProgressBar(None)
            return None
        except:
            self.repos.setProgressBar(None)
            raise

    def setupTs(self):
        return self.doTsSetup()
        
    def setupTsGroup(self):
        return self.doGroupSetup()

    # High-level functions called from YumIO.c

    def collectUpdates(self):
        if yumVersion() >= '2.9.8':
            return self.returnPkgLists([ 'updates' ]).updates
        else:
            self.extcmds = [ 'updates' ]
            return self.returnPkgLists().updates

    def collectAll(self):
        if yumVersion() >= '2.9.8':
            if yumVersion() == '3.2.14':
                # yum 3.2.14 had a bug where an empty argument was interpreted
                # matching no packages instead of all packages.
                return self.returnPkgLists([ 'all', '*' ])
            else:
                return self.returnPkgLists([ 'all' ])
        else:
            return self.returnPkgLists()

    def collectGroupInfo(self):
        try:
            self.doGroupLists(uservisible=1)
        except yum.Errors.GroupsError:
            pass
        if hasDistribGroups():
            return self.comps.groups
        else:
            return self.groupInfo

    def collectDone(self):
        self.closeRpmDB()
        # yum 2.2.2's closeRpmDB will delete the rpmdb, but will leave the
        # localdbimported set to true, so that any future re-created rpmdb
        # won't be populated properly.  Workaround it.
        if hasattr(self, 'localdbimported'):
            self.localdbimported = 0
        self.doTsSetup()

    def pruneObsoletes(self, installs, updates):
        # A set would be nice here, but this must work with python 2.2.x
        # which doesn't support a set.  So use a dict instead.
        obsoleteds = {}
        for insPkg in installs:
            insTuple = (insPkg.name, insPkg.arch,
                        insPkg.epoch, insPkg.version, insPkg.release)
            if insTuple in self.up.obsoletes:
                for (n, a, e, v, r) in self.up.obsoletes[insTuple]:
                    if n != insPkg.name or a != insPkg.arch:
                        obsoleteds[(n, a)] = 1
        for updPair in updates:
            updPkg = updPair[0]
            updTuple = (updPkg.name, updPkg.arch,
                        updPkg.epoch, updPkg.version, updPkg.release)
            if updTuple in self.up.obsoletes:
                for (n, a, e, v, r) in self.up.obsoletes[updTuple]:
                    if n != updPkg.name or a != updPkg.arch:
                        obsoleteds[(n, a)] = 1
        prunedInstalls = []
        prunedUpdates  = []
        for insPkg in installs:
            if (insPkg.name, insPkg.arch) not in obsoleteds:
                prunedInstalls.append(insPkg)
        for updPair in updates:
            updPkg = updPair[0]
            if (updPkg.name, updPkg.arch) not in obsoleteds:
                prunedUpdates.append(updPair)
        return (prunedInstalls, prunedUpdates)
        
    def revertTsInfo(self):
        if yumVersion() >= '3.1.3':
            self.tsInfo = None
            self.tsInfo = self._getTsInfo()
        elif yumVersion() >= '2.5.2':
            self.tsInfo = self._transactionDataFactory()
        else:
            self.tsInfo = yum.transactioninfo.TransactionData()
        
    def apply(self, installs, updates, removes):
        try:
            if self.hasMultiProgress():
                self.repos.setProgressBar(YumIOApplyGrabMeter(),
                                          YumIOApplyGrabMultiMeter())
            else:
                self.repos.setProgressBar(YumIOApplyGrabMeter())
            # Use dsCallback functions from YumIODsCallback
            self.dsCallback = YumIODsCallback(self.log, self.errorlog)
            self.doRpmDBSetup()
            self.doUpdateSetup()
            self.initActionTs()
            (installs, updates) = self.pruneObsoletes(installs, updates)
            for pkg in installs:
                self.tsInfo.addInstall(pkg)
            for pair in updates:
                if hasUpdatedTransactions():
                    self.tsInfo.addUpdate(pair[0], pair[1])
                else:
                    self.tsInfo.addUpdate(pair[0])
            for pkg in removes:
                self.tsInfo.addErase(pkg)
            (result, resultmsgs) = self.buildTransaction()
            if result == 0:
                self.tsInfo = yum.transactioninfo.TransactionData()
                raise Error.YumBaseError, 'Nothing to do'
            elif result == 1:
                self.tsInfo = yum.transactioninfo.TransactionData()
                totalMsg = ''
                for msg in resultmsgs:
                    if totalMsg != '':
                        totalMsg = totalMsg + '\n'
                    totalMsg = totalMsg + msg
                raise yum.Errors.YumBaseError, totalMsg
            elif result == 2:
                # It's good.  Proceed
                pass
            else:
                msg = 'Unexpected result code %u from buildTransaction' % result
                raise Error.YumBaseError, msg
            # This is initialized to RPMPROB_FILTER_FORCERELOCATE |
            # RPMPROB_FILTER_REPLACEOLDFILES (or maybe just to garbage).  So
            # set it to RPMPROB_FILTER_NONE.  This *may* happen automatically
            # because TransactionWrapper.test sets it to
            # RPMPROB_FILTER_DISKSPACE if diskspacecheck is disabled.  But
            # it's best not to rely on that.
            self.ts.ts.setProbFilter(0)
            self.doTransaction()
            self.repos.setProgressBar(None)
            self.revertTsInfo()
            if hasattr(self, 'history'):
                self.history.close()
        except:
            self.repos.setProgressBar(None)
            self.revertTsInfo()
            if hasattr(self, 'history'):
                self.history.close()
            raise

    def applyFailCleanup(self):
        self.closeRpmDB()
        # yum 2.2.2's closeRpmDB will delete the rpmdb, but will leave the
        # localdbimported set to true, so that any future re-created rpmdb
        # won't be populated properly.  Workaround it.
        if hasattr(self, 'localdbimported'):
            self.localdbimported = 0
        self.doTsSetup()

    def collectFileLists(self):
        if self.hasMultiProgress():
            self.repos.setProgressBar(YumIOCollectGrabMeter(),
                                      YumIOCollectGrabMultiMeter())
        else:
            self.repos.setProgressBar(YumIOCollectGrabMeter())
        try:
            self.repos.populateSack('enabled', 'filelists')
            self.repos.setProgressBar(None)
        except:
            self.repos.setProgressBar(None)
            raise

    def collectOtherData(self):
        if self.hasMultiProgress():
            self.repos.setProgressBar(YumIOCollectGrabMeter(),
                                      YumIOCollectGrabMultiMeter())
        else:
            self.repos.setProgressBar(YumIOCollectGrabMeter())
        try:
            self.repos.populateSack('enabled', 'otherdata')
            self.repos.setProgressBar(None)
        except:
            self.repos.setProgressBar(None)
            raise

    def flush(self):
        if not usesPythonLogging():
            if hasattr(self.filelog, 'file_object'):
                self.filelog.file_object.flush()
            elif hasattr(self.filelog, 'list'):
                for log in self.filelog.list:
                    if hasattr(log, 'file_object'):
                        if log.file_object != None:
                            log.file_object.flush()

    def getBiArch(self, arch):
        try:
            return self.biarch[arch]
        except KeyError:
            return arch

    def isBetterThan(self, pkg1, pkg2):
        # Return whether or not the architecture of pkg1 is closer to the
        # base architecture than the architecture of pkg2 is.
        for arch in rpmUtils.arch.getArchList(self.conf.yumvar['basearch']):
            if pkg1.arch == arch:
                return True
            elif pkg2.arch == arch:
                return False
        return True

sys.stdout = YumIOStdout()
sys.stdin = YumIOStdin()
